// Solos PCI driver
// Author : Nathan Williams <nathan@traverse.com.au>
// Last Mod : $Date: 2008-08-25 10:03:43 +1000 (Mon, 25 Aug 2008) $
// Revision : $Rev: 15 $

#define SVNREV "$Rev: 15 $"

#include <linux/version.h> 

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
	#define MODULE
#else
	#include <linux/interrupt.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */
#include <linux/errno.h>	/* error codes */
#include <linux/ioport.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/atm.h>
#include <linux/atmdev.h>
#include <linux/skbuff.h>
#include <linux/sysfs.h>
#include <linux/device.h>
#include <linux/kobject.h>

#include "solos-pci.h"
static int debug = 0;

MODULE_AUTHOR("Traverse Technologies <support@traverse.com.au>");
MODULE_DESCRIPTION("Solos PCI driver");
MODULE_VERSION(VERSION);
MODULE_LICENSE("GPL");
MODULE_PARM_DESC(debug, "Enable Loopback");
module_param(debug, int, 0444);

static struct atm_dev *atmdev[ATM_IF_NUM]; // ATM device pointer
static void __iomem *pciregs;
static void __iomem *pciregs0;
static u8 myirq;
static int opens;
static struct sk_buff_head tx_queue[4];
static struct sk_buff_head cli_queue[4];

static int send_fpga(void);
static irqreturn_t short_interrupt(int irq, void *dev_id);
static struct atm_vcc* find_vcc(struct atm_dev *dev, short vpi, int vci);
static int list_vccs(int vci);
static int atm_init(void);
static void atm_remove(void);
static void kobj_release(struct kobject *kobj);
static ssize_t sysfs_show(struct kobject *kobj, struct attribute *attr, char *buf);
static ssize_t sysfs_store(struct kobject *kobj, struct attribute *attr, const char *buf, size_t count);
static int send_command(int dev, const char *buf);
static void interrupt_bh(unsigned long);

DECLARE_TASKLET(isr_bh,interrupt_bh,0);

static struct kobject *solos_kobj;

static spinlock_t tx_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t tx_queue_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t cli_queue_lock = SPIN_LOCK_UNLOCKED;
static unsigned long irq_flags;
static unsigned long irq_flags2;
static int bh_count = 0;

static ssize_t sysfs_show(struct kobject *kobj, struct attribute *attr,
      char *buf)
{
	struct sk_buff *skb=NULL;

	if(strcmp(attr->name,"stats") == 0){
	  return sprintf(buf, "This feature has not been implemented.\n");
	} else if(strcmp(attr->name,"console0") == 0){
			if(ATM_IF_NUM <1){
			  return sprintf(buf, "This device is disabled.\n");
			}
  		spin_lock(&cli_queue_lock);
 	 		skb = skb_dequeue(&cli_queue[0]);
  		spin_unlock(&cli_queue_lock);
			if(skb == NULL){
			  return sprintf(buf, "No data.\n");
			}
			memcpy(buf, skb->data, skb->len);
			PDEBUG("len: %d\n",skb->len);
			kfree_skb(skb);
			return skb->len;
				
	} else if(strcmp(attr->name,"console1") == 0){
			if(ATM_IF_NUM <2){
			  return sprintf(buf, "This device is disabled.\n");
			}
	} else if(strcmp(attr->name,"console2") == 0){
			if(ATM_IF_NUM <3){
			  return sprintf(buf, "This device is disabled.\n");
			}
	} else if(strcmp(attr->name,"console3") == 0){
			if(ATM_IF_NUM <4){
			  return sprintf(buf, "This device is disabled.\n");
			}
	}
	return 0;
}

static int send_command(int dev, const char *buf){
  struct sk_buff *skb=NULL;
	int size;
  unsigned char *header=NULL;

	size = strlen(buf);
//	PDEBUG("size: %d\n", size);
	if(size > (BUF_SIZE - HEADER_SIZE )){
		PDEBUG("Command is too big.  Dropping request\n");
		return 0;
	}
	skb = alloc_skb(size+HEADER_SIZE,GFP_ATOMIC);
	if(skb == NULL){
		PDEBUG("Failed to allocate sk_buff in send_command()\n");
		printk(KERN_WARNING "%s: Failed to send command.\n", PTAG);
		return 0;
	} 

  header = skb_put(skb, HEADER_SIZE);
	if(header == NULL){
		PDEBUG("Failed to push header in send_command()\n");
		return 0;
	}

  *header =     (size & 0x00FF);
  *(header + 1)=(size & 0xFF00) >> 8;
  *(header + 2)=0;//Reserved VPI
  *(header + 3)=0;//Reserved VPI
  *(header + 4)=0;//Reserved VCI
  *(header + 5)=0;//Reserved VCI
  *(header + 6)=1;//Type
  *(header + 7)=0;

	memcpy(skb_put(skb,size), buf,size);

  spin_lock(&tx_queue_lock);
  skb_queue_tail(&tx_queue[dev],skb);
  spin_unlock(&tx_queue_lock);
	send_fpga();

	return 0;
}

static ssize_t sysfs_store(struct kobject *kobj, struct attribute *attr,
       const char *buf, size_t count)
{
	if(strcmp(attr->name,"stats") == 0){
		return count;
	} else if(strcmp(attr->name,"console0") == 0){
		send_command(0,buf);
	} else if(strcmp(attr->name,"console1") == 0){
		send_command(1,buf);
	} else if(strcmp(attr->name,"console2") == 0){
		send_command(2,buf);
	} else if(strcmp(attr->name,"console3") == 0){
		send_command(3,buf);
	}
	return count;
}

static void kobj_release(struct kobject *kobj)
{
	kfree(kobj);
}

static struct sysfs_ops kobj_sysfs_ops = {
	.show = sysfs_show,
	.store = sysfs_store
};

#define DEF_ATTR(_obj,_name,_mode) \
	static struct attribute _obj = { .name = _name, .mode = _mode }

DEF_ATTR(file1_attr, "stats", 0444);
DEF_ATTR(file2_attr, "console0", 0644);
DEF_ATTR(file3_attr, "console1", 0644);
DEF_ATTR(file4_attr, "console2", 0644);
DEF_ATTR(file5_attr, "console3", 0644);

static struct attribute *kobj_default_attrs[] = {
  &file1_attr,
  &file2_attr,
  &file3_attr,
  &file4_attr,
  &file5_attr,
  NULL, /* need to NULL terminate the list of attributes */
};

static struct kobj_type kobj_solos = {
	.release = kobj_release,
	.sysfs_ops = &kobj_sysfs_ops,
	.default_attrs = kobj_default_attrs
};





static irqreturn_t short_interrupt(int irq, void *dev_id)
{
	int handled = 1;

//ACK IRQ
	iowrite32(0x00,pciregs0 + IRQ_CLEAR);
//Disable IRQs from FPGA	
	iowrite32(0x00,pciregs0 + IRQ_EN_ADDR);
	if(opens != 0){
		tasklet_schedule(&isr_bh);
	}

	bh_count++;
//Enable IRQs from FPGA	
	iowrite32(0x01,pciregs0 + IRQ_EN_ADDR);
	return IRQ_RETVAL(handled);
}

void interrupt_bh(unsigned long unused)
{
	int size,dev_num,vci;
	short vpi;
	struct sk_buff *skb;
	unsigned char txbyte, rxbyte;
	struct atm_vcc *vcc;
	int type;
	u32 data32;
	u32 rxflags;
	bh_count=0;

	data32 = ioread32(pciregs0 + FLAGS_ADDR); //Read flags
	txbyte = (unsigned char)(0x0000000F & data32);
//PDEBUG("Interrupting!\n");
	rxbyte = (unsigned char)(0x000000F0 & data32)>>4;

//Now mask out unused ports so they are set to '1'
//e.g. ATM_IF_NUM = 2 then txbyte becomes 000011XX (binary)
	txbyte |= (0x0F & (0x0F << ATM_IF_NUM)); //mask unused ports
//PDEBUG("RX full: %X TX empty: %X\n",rxbyte,txbyte);

	if(txbyte != 0x0F){ //TX buf empty
		send_fpga();
	} else {
//		PDEBUG("TX busy\n");
	}

	rxflags = 0;
	for(dev_num=0;dev_num<4;dev_num++){
		if(((rxbyte >> dev_num) & 0x01) == 0x01){
			rxflags |= (u32)(0x00000010) << dev_num;
//		rxflags = (u32)(0x00000010) << dev_num;
			data32 =ioread32(pciregs + (dev_num * MODEM_SIZE) + RX_BUF_ADDR );
			size =  (int)(0x000000FF & data32);
			size |= (int)(0x0000FF00 & data32);
			vpi =   (int)(0x00FF0000 & data32) >> 16;
			vpi |=  (int)(0xFF000000 & data32) >> 24;
			data32 =ioread32(pciregs + (dev_num * MODEM_SIZE) + RX_BUF_ADDR + 4);
//		PDEBUG("data32 is %08lX\n",(unsigned long)data32);

			vci =  (0x000000FF & data32);
			vci |= (0x0000FF00 & data32);
			type = (0x00FF0000 & data32) >> 16; //type of data
//		PDEBUG("size: %d VPI: %d VCI: %d\n",size,vpi,vci);
//		PDEBUG("Fetching %d bytes for device %d\n",size,dev_num);
		
			skb = alloc_skb(size+3,GFP_ATOMIC);//+3 allocated in case size isn't divisible by 4
			if(skb == NULL){
				PDEBUG("%s: Failed to allocate sk_buff in short_interrupt()\n", PTAG);
				printk(KERN_WARNING "%s: Lost packet!!\n", PTAG);
		  	continue;
			} 
	
			memcpy(skb_put(skb,size), pciregs + (dev_num * MODEM_SIZE) + HEADER_SIZE + RX_BUF_ADDR,size);

			skb_trim(skb,size); //Trim down to size

			if(type == 0){
				vcc = find_vcc(atmdev[dev_num],vpi,vci); 
				if(vcc == NULL){
//				PDEBUG("vcc is NULL\n");
					continue;
				} 
				atm_charge(vcc,skb->truesize);
				vcc->push(vcc,skb);	
				atomic_inc(&vcc->stats->rx);
			} else {
      	spin_lock(&cli_queue_lock);
	      skb_queue_tail(&cli_queue[dev_num],skb);
	      spin_unlock(&cli_queue_lock);
			}
		}
	}
	if(rxflags != 0){
		iowrite32(rxflags,pciregs0 + FLAGS_ADDR);
	}
	return;
}

static struct atm_vcc* find_vcc(struct atm_dev *dev, short vpi, int vci)
{
	struct hlist_head *head;
	struct atm_vcc *vcc;
	struct hlist_node *node;
	struct sock *s;

			 head = &vcc_hash[vci & (VCC_HTABLE_SIZE -1)];

			 sk_for_each(s, node, head) {
							 vcc = atm_sk(s);
							 if (vcc->dev == dev &&
									 vcc->vci == vci && vcc->vpi == vpi &&
									 vcc->qos.rxtp.traffic_class != ATM_NONE) {
															 return vcc;
							 }
			 }
			 return NULL;
}

static int list_vccs(int vci){
	struct hlist_head *head;
	struct atm_vcc *vcc;
	struct hlist_node *node;
	struct sock *s;
	int num_found = 0;
	int i;

	if (vci != 0){
		head = &vcc_hash[vci & (VCC_HTABLE_SIZE -1)];
		sk_for_each(s, node, head) {
			num_found ++;
			vcc = atm_sk(s);
			printk(KERN_DEBUG "Device: %d Vpi: %d Vci: %d\n", vcc->dev->number, vcc->vpi, vcc->vci);
		}
	} else {
		for(i=0;i<32;i++){
			head = &vcc_hash[i];
			sk_for_each(s, node, head) {
				num_found ++;
	 			vcc = atm_sk(s);
				printk(KERN_DEBUG "Device: %d Vpi: %d Vci: %d\n", vcc->dev->number, vcc->vpi, vcc->vci);
		 	}
		}
	}
	return num_found;
}


static int popen(struct atm_vcc *vcc){
	struct sk_buff *skb=NULL;
	unsigned char *header=NULL;
	skb = alloc_skb(HEADER_SIZE,GFP_ATOMIC);
	if(skb == NULL){
		PDEBUG("Failed to allocate sk_buff in popen()\n");
		printk(KERN_WARNING "%s: Failed to open vcc.\n", PTAG);
		return 0;
	} 
	header = skb_put(skb, HEADER_SIZE);
	if(header == NULL){
		PDEBUG("Failed to push header in send_command()\n");
		return 0;
	}
  *header =     (HEADER_SIZE & 0x00FF);
  *(header + 1)=(HEADER_SIZE & 0xFF00) >> 8;
	*(header + 2)=(vcc->vpi & 0x00FF);
	*(header + 3)=(vcc->vpi & 0xFF00) >> 8;
	*(header + 4)=(vcc->vci & 0x00FF);
	*(header + 5)=(vcc->vci & 0xFF00) >> 8;
  *(header + 6)=3;//Type is popen
  *(header + 7)=0;
 	
	spin_lock(&tx_queue_lock);
  skb_queue_tail(&tx_queue[vcc->dev->number],skb);
  spin_unlock(&tx_queue_lock);
	send_fpga();

//	PDEBUG("Open for vpi %d and vci %d on interface %d\n", vcc->vpi, vcc->vci, vcc->dev->number);
	set_bit(ATM_VF_ADDR, &vcc->flags); // accept the vpi / vci
	set_bit(ATM_VF_READY, &vcc->flags);
	list_vccs(0);
	if(opens ==0){
		iowrite32(0x01,pciregs0 + IRQ_EN_ADDR);
	}
	opens = opens + 1; //count open PVCs
	return(0);
}

static void pclose(struct atm_vcc *vcc){
	struct sk_buff *skb=NULL;
	unsigned char *header=NULL;
	skb = alloc_skb(HEADER_SIZE,GFP_ATOMIC);
	if(skb == NULL){
		PDEBUG("Failed to allocate sk_buff in popen()\n");
		printk(KERN_WARNING "%s: Failed to open vcc.\n", PTAG);
		return;
	} 
	header = skb_put(skb, HEADER_SIZE);
	if(header == NULL){
		PDEBUG("Failed to push header in send_command()\n");
		return;
	}
  *header =     (HEADER_SIZE & 0x00FF);
  *(header + 1)=(HEADER_SIZE & 0xFF00) >> 8;
	*(header + 2)=(vcc->vpi & 0x00FF);
	*(header + 3)=(vcc->vpi & 0xFF00) >> 8;
	*(header + 4)=(vcc->vci & 0x00FF);
	*(header + 5)=(vcc->vci & 0xFF00) >> 8;
  *(header + 6)=4;//Type is pclose
  *(header + 7)=0;
 	
	spin_lock(&tx_queue_lock);
  skb_queue_tail(&tx_queue[vcc->dev->number],skb);
  spin_unlock(&tx_queue_lock);
	send_fpga();
//	PDEBUG("Close for vpi %d and vci %d on interface %d\n", vcc->vpi, vcc->vci, vcc->dev->number);
	opens = opens - 1;
	if(opens ==0){
		iowrite32(0x00,pciregs0 + IRQ_EN_ADDR);
	}
	clear_bit(ATM_VF_ADDR, &vcc->flags); 
	clear_bit(ATM_VF_READY, &vcc->flags); 
	return;
}

static int send_fpga(void){
	int len,i,j;
	u32 data32;
	u32 txflag;
	char byte;
	struct sk_buff *buf=NULL;
	unsigned char device;
	spin_lock_irqsave(&tx_lock,irq_flags);
	txflag = ioread32(pciregs0 + FLAGS_ADDR); //Read flags
	byte = (unsigned char)(txflag & 0x0000000F); //mask out non-txflag bits
	byte |= (0x0F & (0x0F << ATM_IF_NUM)); //mask unused ports
	txflag = 0; //clear so we can set bits later
	//printk(KERN_INFO "TX Flags are %X\n",byte);

	for(device=0;device<4;device++){
//	PDEBUG("Checking modem %d:  %X\n",id,	((byte >> id) & 0x01));
		if(((byte >> device) & 0x01) == 0){
			spin_lock_irqsave(&tx_queue_lock,irq_flags2);
			buf = skb_dequeue(&tx_queue[device]);
			spin_unlock_irqrestore(&tx_queue_lock,irq_flags2);
			if(buf !=NULL){
				len = buf->len;
				for(i=0;i<(len-(len%4));i+=4){
					data32 = 0;
					data32 |= (u32)*(buf->data + i);
					data32 |= (u32)*(buf->data + i + 1) << 8;
					data32 |= (u32)*(buf->data + i + 2) << 16;
					data32 |= (u32)*(buf->data + i + 3) << 24;
					iowrite32(data32,pciregs+(i)+ (device * MODEM_SIZE) + TX_BUF_ADDR );
				}
				data32 = 0;
				for(j=0;j<(len%4);j++){
					data32 |= (u32)*(buf->data + i + j) << (8*j);
				}
				kfree_skb(buf);
//			PDEBUG("%08lX\n",(unsigned long)data32);
				if(len%4 > 0){
					iowrite32(data32,pciregs+(i)+ (device * MODEM_SIZE) + TX_BUF_ADDR );
				}
				txflag |= (0x00000001 << device);//Set TX full flag
			}
		}
	}
	iowrite32(txflag,pciregs0 + FLAGS_ADDR); //Write word
//PDEBUG("Queue %d %d %d %d\n",skb_queue_len(&tx_queue[0]),skb_queue_len(&tx_queue[1]),skb_queue_len(&tx_queue[2]),skb_queue_len(&tx_queue[3]));
	spin_unlock_irqrestore(&tx_lock,irq_flags);
	return 0;
}

static int psend(struct atm_vcc *vcc,struct sk_buff *skb){
	struct atm_vcc *test;
	struct sk_buff *skb2=NULL;
	struct sk_buff *skbcopy=NULL;
	unsigned char *header=NULL;
//PDEBUG("psend called.\n");
//PDEBUG("dev,vpi,vci = %d,%d,%d\n",vcc->dev->number,vcc->vpi,vcc->vci);
	if((skb->len > (BUF_SIZE - HEADER_SIZE )) && (debug ==0)){
		printk(KERN_INFO "Length of PDU is too large. Dropping PDU.\n");
		if(vcc->pop == NULL){
			printk(KERN_DEBUG "Using dev_kfree_skb because vcc->pop is NULL\n");
			dev_kfree_skb(skb);
		} else{
			vcc->pop(vcc, skb);
		}
		return 0;
	}

	test = find_vcc(vcc->dev, vcc->vpi, vcc->vci);
	if(test){
		if(debug==0){
			skbcopy = skb_copy_expand(skb,HEADER_SIZE,0,GFP_ATOMIC);
			header = skb_push(skbcopy, HEADER_SIZE);
			*header =     (skb->len & 0x00FF);
			*(header + 1)=(skb->len & 0xFF00) >> 8;
			*(header + 2)=(vcc->vpi & 0x00FF);
			*(header + 3)=(vcc->vpi & 0xFF00) >> 8;
			*(header + 4)=(vcc->vci & 0x00FF);
			*(header + 5)=(vcc->vci & 0xFF00) >> 8;
			*(header + 6)=0;//Type field (0 for data)
			*(header + 7)=0;
		//	skb_put(skbcopy, HEADER_SIZE);
			spin_lock(&tx_queue_lock);
			skb_queue_tail(&tx_queue[vcc->dev->number],skbcopy);
			spin_unlock(&tx_queue_lock);
//			PDEBUG("Queue %d %d %d %d\n",skb_queue_len(&tx_queue[0]),skb_queue_len(&tx_queue[1]),skb_queue_len(&tx_queue[2]),skb_queue_len(&tx_queue[3]));
			send_fpga();
			atomic_inc(&vcc->stats->tx) ;
			if(vcc->pop == NULL){
				printk(KERN_DEBUG "Using dev_kfree_skb because vcc->pop is NULL\n");
				dev_kfree_skb(skb);
			} else{
		    vcc->pop(vcc, skb);
		  }
			return 0;
		} else {	
			skb2=atm_alloc_charge(test, skb->len, GFP_ATOMIC);
			if(skb2 != NULL){
				memcpy(skb2->data,skb->data,skb->len);
				skb_put(skb2,skb->len);
			}
			test->push(test,skb2);	
			atomic_inc(&test->stats->rx)	;
			if(vcc->pop == NULL){
				printk(KERN_DEBUG "Using dev_kfree_skb because vcc->pop is NULL\n");
				dev_kfree_skb(skb);
			} else{
				vcc->pop(vcc, skb);
				atomic_inc(&vcc->stats->tx)	;
			} 
		}
	} else {
		PDEBUG("Failed to find matching VCC\n");
		if(vcc->pop == NULL){
			printk(KERN_DEBUG "Using dev_kfree_skb because vcc->pop is NULL\n");
			dev_kfree_skb(skb);
		} else{
			vcc->pop(vcc, skb);
		} 
	}
	return(0);
}

static struct atmdev_ops fpga_ops = {
	.open =		popen,
	.close =	pclose,
	.ioctl =	NULL,
	.getsockopt =	NULL,
	.setsockopt =	NULL,
	.send =		psend,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
	.sg_send =	NULL,
	.feedback =	NULL,
#endif
	.send_oam =	NULL,
	.phy_put =	NULL,
	.phy_get =	NULL,
	.change_qos =	NULL,
	.proc_read =	NULL,
	.owner =	THIS_MODULE
};

static int fpga_probe(struct pci_dev *dev, const struct pci_device_id *id) {
	int err, i;
	u16 fpga_ver;
	u8 major_ver;
	u8 minor_ver;
	unsigned long addr_start, addr_end, addr_len;
	unsigned long caddr_start, caddr_end, caddr_len; //config BAR0
	u8 pci_bar,pci_vend;
	u32 data32;
	

if (debug == 0){
	err = pci_enable_device(dev);
	if(err){
		printk(KERN_ALERT "Failed to enable PCI device\n");
	}
	
	err = pci_read_config_byte(dev, PCI_VENDOR_ID, &pci_vend);
	if(err){
		printk(KERN_ALERT "Failed to read PCI_VENDOR_ID\n");
	}
	
	err = pci_read_config_byte(dev, PCI_BASE_ADDRESS_1, &pci_bar);
	if(err){
		printk(KERN_ALERT "Failed to read PCI_BASE_ADDRESS_1\n");
	}
	
//	err = pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &myirq);
//	if(err){
//		printk(KERN_ALERT "Failed to read PCI_INTERRRUPT_LINE\n");
//	}
//	PDEBUG("IRQ : %d\n", myirq);

	caddr_start = pci_resource_start(dev,0); //BAR0 start address
	caddr_end	 = pci_resource_end(dev,0); //BAR0 end address
	caddr_len	 = pci_resource_len(dev,0);
	addr_start = pci_resource_start(dev,1); //BAR1 start address
	addr_end	 = pci_resource_end(dev,1); //BAR1 end address
	addr_len	 = pci_resource_len(dev,1);

	printk(KERN_DEBUG "BAR0 Start Address: %lX End Address : %lX\n", caddr_start, caddr_end);
	printk(KERN_DEBUG "Length : %lu 32-bit words\n", caddr_len);
	printk(KERN_DEBUG "BAR1 Start Address: %lX End Address : %lX\n", addr_start, addr_end);
	printk(KERN_DEBUG "Length : %lu 32-bit words\n", addr_len);

	err = pci_request_region(dev,0,"name");
	if(err != 0){
	printk(KERN_ALERT "Failed to request PCI region\n");
	}
	request_mem_region(caddr_start,caddr_len,"name");
	err = pci_request_region(dev,1,"name");
	if(err != 0){
	printk(KERN_ALERT "Failed to request PCI region\n");
	}
	request_mem_region(addr_start,addr_len,"name");
	
	pciregs0 = ioremap(caddr_start, caddr_len);
	pciregs = ioremap(addr_start, addr_len);



//	for(i=0;i<64 ;i+=4){
//		data32=ioread32(pciregs + i);
//		PDEBUG("%08lX\n",(unsigned long)data32);
//	}
//Fill Config Mem with zeros
	for(i=0;i<128;i+=4){
		data32 = 0x00000000;
		iowrite32(data32,pciregs0+i);
	}

		data32 = 0x000000F0; //Setting RX empty flag
		iowrite32(data32,pciregs0+FLAGS_ADDR);

	data32 = ioread32(pciregs0 + FPGA_VER);
	fpga_ver = (data32 & 0x0000FFFF);
	major_ver = ((data32 & 0xFF000000) >> 24);
	minor_ver = ((data32 & 0x00FF0000) >> 16);
	printk(KERN_INFO "FPGA Version %d.%02d svn-%d\n",major_ver, minor_ver,fpga_ver);

//Set Loopback mode
//		data32 = 0x00010000;
//		iowrite32(data32,pciregs0 + FLAGS_ADDR);

//	Enable IRQs
		data32 = 0x00000001;
		iowrite32(data32,pciregs0 + IRQ_EN_ADDR);


	//Fill Buffers with zeros
//	for(i=0;i<BUF_SIZE *8 ;i+=4){
//		data32 = 0x00000000;
//		iowrite32(data32,pciregs+i);
//	}




//	for(i=0;i<(BUF_SIZE *1) ;i+=4){
//		data32 = 0x12345678;
//		iowrite32(data32,pciregs+i + (0*BUF_SIZE));
//	}
//	for(i=0;i<(BUF_SIZE *1) ;i+=4){
//		data32 = 0xabcdef98;
//		iowrite32(data32,pciregs+i + (1*BUF_SIZE));
//	}

	//Read Config Memory
//	printk(KERN_DEBUG "Reading Config MEM\n");
//	i=0;
//	for(i=0;i<16;i++){
//		data32=ioread32(pciregs + i*(BUF_SIZE/2));
	//	printk(KERN_ALERT "Addr: %lX Data: %08lX\n",(unsigned long)(addr_start + i*(BUF_SIZE/2)),(unsigned long)data32);
//	}
myirq = dev->irq;
//PDEBUG("Requesting IRQ: %d\n",dev->irq);
	err = request_irq(dev->irq, short_interrupt, IRQF_DISABLED|IRQF_SHARED, "solos-pci", pciregs);
	if(err){
		PDEBUG("Failed to request interrupt IRQ: %d\n",myirq);
	}
}
	return(0);
}

static int atm_init(void) {
	int i;
	int err = 0;
	
	opens = 0;

  for(i=0;i<ATM_IF_NUM;i++){
    atmdev[i] = atm_dev_register("solos-pci", &fpga_ops, -1, NULL);
    if (!atmdev[i]){
      printk(KERN_ALERT "%s: Could not register ATM device %d\n", PTAG, i);
      err = -ENODEV;
    }
    printk(KERN_INFO "Registered ATM device %d\n", atmdev[i]->number);
    if (!atmdev[i]->ops){
      printk(KERN_ALERT "%s: No ATM device ops\n", PTAG);
    }
    atmdev[i]->ci_range.vpi_bits = 8;
    atmdev[i]->ci_range.vci_bits = 16;
    atmdev[i]->dev_data = NULL;
  }
  //initialise queue
  skb_queue_head_init(&tx_queue[0]);
  skb_queue_head_init(&tx_queue[1]);
  skb_queue_head_init(&tx_queue[2]);
  skb_queue_head_init(&tx_queue[3]);
  skb_queue_head_init(&cli_queue[0]);
  skb_queue_head_init(&cli_queue[1]);
  skb_queue_head_init(&cli_queue[2]);
  skb_queue_head_init(&cli_queue[3]);
	return err;
}

static void atm_remove(void){
	int i;
	for(i=0;i<ATM_IF_NUM;i++){
	printk(KERN_INFO "Unregistering ATM device %d\n",atmdev[i]->number);
		atm_dev_deregister(atmdev[i]);
	}
	return;
}


static void fpga_remove(struct pci_dev *dev) {
	unsigned long addr_start, addr_end, addr_len;
if(debug ==0){
	addr_start = pci_resource_start(dev,1); //BAR0 start address
	addr_end	 = pci_resource_end(dev,1); //BAR0 end address
	addr_len	 = pci_resource_len(dev,1);
}
if(debug ==0){	
	printk(KERN_DEBUG "Freeing IRQ\n");
// Disable IRQs from FPGA	
	iowrite32(0x00,pciregs0 + IRQ_EN_ADDR);
	free_irq(myirq, pciregs);
//	iowrite32(0x01,pciregs);
	printk(KERN_DEBUG "Unmapping PCI resource\n");
	iounmap(&pci_resource_start(dev,0));
	iounmap(&pci_resource_start(dev,1));
	printk(KERN_DEBUG "Releasing PCI Region\n");
	pci_release_region(dev,0);
	pci_release_region(dev,1);
//	release_mem_region(pci_resource_start(dev, 1), pci_resource_len(dev, 1));
	pci_disable_device(dev);

//	PDEBUG("fpga_remove\n");
	}
	return;
}

static struct pci_device_id fpga_pci_tbl[] __devinitdata = {
	{ 0x10ee, 0x0300, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0, }
};

MODULE_DEVICE_TABLE(pci,fpga_pci_tbl);

static struct pci_driver fpga_driver = {
	.name =		"pci_skel",
	.id_table =	fpga_pci_tbl,
	.probe =	fpga_probe,
	.remove =	fpga_remove,
};


static int __init pci_skel_init(void)
{
	int err;
	int i;
	char svnrev[16]=SVNREV;
	struct module *me = THIS_MODULE;
	for(i=6;i<16;i++){
		if (svnrev[i] == ' '){ //space after the revision number
			svnrev[i] = 0;
		}
	}
	strcpy(svnrev, (svnrev + 6)); //skip over "Rev: "


	err = atm_init();
  if(err){
		return err;
	}
	printk(KERN_INFO "Solos PCI Driver Version %s svn-%s\n", VERSION, svnrev);
	/////example_kobj = kobject_create_and_add("kobject_example", &((dev->dev).kobj));
//	example_kobj = kobject_create_and_add("kobject_example", &(me->mkobj.kobj) );
  solos_kobj = kzalloc(sizeof(*solos_kobj),GFP_KERNEL);
	if(solos_kobj == NULL){
		return -ENOMEM;
	}
	kobject_set_name(solos_kobj,"%s","interfaces");
	solos_kobj->parent = &(me->mkobj.kobj);
	solos_kobj->ktype = &kobj_solos;
	err = kobject_register(solos_kobj);
  if (err){
		PDEBUG("Error registering kobject\n");
    kobject_put(solos_kobj);
	}

	return pci_register_driver(&fpga_driver);
}

static void __exit pci_skel_exit(void)
{
	kobject_put(solos_kobj);
	atm_remove();
	pci_unregister_driver(&fpga_driver);
	printk(KERN_INFO "Solos PCI Driver %s Unloaded\n", VERSION);
}

module_init(pci_skel_init);
module_exit(pci_skel_exit);
